#Helper functions

_processRevisions = (instructionName, config, collectionName, revisions) ->
    #are the parameters consistent with the config?
    if not config.revisions? or not config.revisions[collectionName]?
        return [new Error("can't #{instructionName} documents in #{collectionName} without relevant config.revisions supplied at committed.start")]
    #revisions can be null, string, array, or object
    #if revisions is null, then grab all the revisions from config and use these
    if not revisions? then revisions = config.revisions[collectionName]
    #if it's a string, then make it an array
    if typeof revisions is 'string' then revisions = [revisions]
    if Array.isArray(revisions)
        #if we now have an array then we havn't been given specific version numbers, so these are null
        revisionNames = revisions
        revisionNumbers = null
    else
        #if we were given an object with exact revision numbers, make a names list, and return both
        revisionNames = ( key for key of revisions )
        revisionNumbers = revisions
    #now check the config is consistent with the revision names
    for revisionName in revisionNames
        if revisionName not in config.revisions[collectionName]
            return [new Error("can't #{instructionName} documents with a revision '#{revisionName}' not specified in the config.revisions given at committed.start")]
    return [null, revisionNames, revisionNumbers]


# insert instructions

exports.insert = 
    preSave: (config, transaction, state, [collectionName, documents, revisions, etc...], done) ->
        if etc.length isnt 0 then return done(new Error("too many values passed to insert"))
        if not (collectionName? and documents?) then return done(new Error("null or missing argument to insert command"))
        if not Array.isArray(documents) then documents = [documents]

        [error, revisionNames, revisionNumbers] = _processRevisions 'insert', config, collectionName, revisions
        if error?
            return done(error)
        if revisionNumbers? then return done(new Error("can't insert using specific revision numbers #{revisionNumbers}"))

        #check the docs dont have any revision info if we're intending to automatically set revision info
        if documents.some((d) -> d.revision?)
            return done(new Error("can't insert documents when some already have a revision property"))

        config.db.collection collectionName, {strict: true}, (err, collection) ->
            if err? then return done(err)
            #inserts should _id every document, otherwise accurate rollback won't be possible, moreover these ids need to be saved somewhere
            state.ids = []
            for d in documents
                if d._id? 
                    state.ids.push d._id
                else
                    d._id = _pkFactory.createPk()
                    state.ids.push d._id
                #while we're looking at documents, check to see if we should create some revision info
                if revisionNames.length > 0 and not d.revision?
                    d.revision = {}
                    for revisionName in revisionNames
                        d.revision[revisionName] = 0
            done()

    execute: (config, transaction, state, [collectionName, documents, revisions, etc...], done) ->
        options =
                w:1
                journal: true
                serializeFunctions: false
        collection.insert documents, options, (err, objects) ->
            if err? then return done(err, null)
            return done(null, true)

exports.insertRollback = 
    preSave: (config, transaction, state, [collectionName, documents, revisions, etc...], done) ->
        if etc.length isnt 0 then return done(new Error("too many values passed to insertRollback"))
        if not (collectionName? and documents?) then return done(new Error("null or missing argument to insertRollback command"))
        done()

    execute: (config, transaction, state, [collectionName, documents, revisions, etc...], done) ->
        config.db.collection collectionName, {strict: true}, (err, collection) ->
            if err? then return done(err, false)
            options =
                w:1
                journal: true
                serializeFunctions: false
            if not state.ids?
                #then we never got to the insert, so there's nothing for the rollback to do
                return done(null, true)
            else
                #where documents lack an _id, we have to fill this in from what was generated during the insert instruction. 
                if not Array.isArray(documents) then documents = [documents]
                if documents.length isnt state.ids.length
                    return done(new Error("during rollback, number of documents to insert doesnt match the number of ids generated during insert"), false)
                for d,i in documents
                    if not d._id? then d._id = state.ids[i]
                iterator = (item, itemDone) ->
                    #remove justOne item, but only if it's an exact match for the inserted document
                    collection.remove item, options, (err, result) ->
                        itemDone(err)
                return async.each documents, iterator, (err) ->
                    return done(err, true) #rollbacks only fail if there's an error

#Update instructions

_updateOneOptions =
    w:1
    journal: true
    upsert:false
    multi: false
    serializeFunctions: false

_updateManyOptions = 
    w: 1
    journal: true
    upsert: false
    multi: true
    serializeFunctions: false

_mongoOpsFromTo = (mongoOps, fromDoc, toDoc) ->
    toFlattened = _mongoFlatten(toDoc)
    fromFlattened = _mongoFlatten(fromDoc)
    #the documents could contain too many revision keys, so don't copy these. 
    mongoOps.$set[key] = value for [key, value] in toFlattened when key.indexOf('revision.') isnt 0
    #make a list of fields to unset
    newDotPaths = ( key for key, value of mongoOps.$set )
    oldDotPaths = ( key for [key, value] in fromFlattened when key.indexOf('revision.') isnt 0 )
    unsetDotPaths = ( path for path in oldDotPaths when path not in newDotPaths )
    for path in unsetDotPaths
        mongoOps.$unset[path] = 1
    #remove _id from $set, we dont want to update this
    delete mongoOps.$set._id

_updateOpInstruction = (onlyOne, config, transaction, state, collectionName, selector, updateOps, rollbackProjector, revisions, etc, done) ->
    instructionName = if onlyOne then 'updateOneOp' else 'updateManyOp'
    if etc.length isnt 0 then return done(new Error("too many values passed to #{instructionName}"))
    if not (collectionName? and selector? and updateOps?) then return done(new Error("null or missing argument to #{instructionName} instruction"))

    [error, revisionNames, revisionNumbers] = _processRevisions instructionName, config, collectionName, revisions
    if error?
        return done(error)

    if not onlyOne and revisionNumbers?
        return done( new Error("can't #{instructionName} with a revisions object that specifies exact revision numbers, that's only possible for singe updates"))
    #can't proceed if we've already got an $inc updateOps that involves the revisions 
    if updateOps.__inc? and revisionNames.some((name) -> updateOps.__inc["revision.#{name}"])
        return done(new Error("can't #{instructionName} with revisions #{revisionNames} when there are already $inc operations on the revision sub-document"))
    #sames true of the selector if we're going to match an exact version number, this isnt really that precise a check, but it's at least a basic check
    if revisionNumbers? and revisionNames.some((name) -> selector["revision.#{name}"])
        return done(new Error("can't #{instructionName} with revisions #{revisionNumbers} with a selector that already matches on the revision subdoc"))

    config.db.collection collectionName, {strict: true}, (err, collection) ->
        if err? then return done(err, false)

        #irrespective of whether we've been given required revision numbers, we need their current values written into the transaction right away
        mongoSelector = _mongolize(selector)
        if rollbackProjector?
            mongoProjector = _mongolize(rollbackProjector)
            mongoProjector._id = 1
            #always get the revision sub-doc - whittle this down to just the revisions of interest during rollback
            mongoProjector.revision = 1
        else
            #if we weren't given a rollback projector then grab the whole document.
            mongoProjector = {}

        #we're going to read current values into memory, then send it back out
        #to the transaction collection. this instruction consequently needs to
        #be used with care, if you need to update a million documents,
        #consider using a different instruction.
        collection.find(mongoSelector, mongoProjector).toArray (err, docs) ->
            if err? then return done(err, null)
            if onlyOne and docs.length isnt 1 
                #the instruction has failed, we must find exactly one doc
                transaction.execution.info.push "#{instructionName} can update only one document, using #{JSON.stringify mongoSelector} #{docs.length} were found"
                return done(null, false) 
            #if we have a document to update, we should save its id and exact revision into the transaction
            if onlyOne
                state.updated = docs[0]
            else
                state.updated = docs
            _updateTransactionState transaction, (err) ->
                if err? then return done(err, null)
                #now we're safe for a rollback 

                executeUpdate = () ->
                    if onlyOne then options = _updateOneOptions else options = _updateManyOptions
                    mongoUpdateOps = _mongolize(updateOps)
                    if revisionNames.length > 0
                        if not mongoUpdateOps.$inc? then mongoUpdateOps.$inc = {}
                        mongoUpdateOps.$inc["revision.#{revisionName}"] = 1 for revisionName in revisionNames
                    #note that we're not using $isolated. This doesn't work
                    #over shards and committed's queue structure should be
                    #used to ensure isolation.
                    collection.update mongoSelector, mongoUpdateOps, options, (err, updated) ->
                        if err? then return done(err, null)
                        if onlyOne and updated isnt 1 
                            transaction.execution.info.push """
                                #{instructionName} can update only one document, using #{JSON.stringify mongoSelector} #{updated} were updated
                                """
                            return done(null, false) #the instruction has failed
                        if onlyOne
                            return done(null, true)
                        else
                            #if this is an update many then return the number of records updated.
                            return done(null, true, updated)

                if revisionNumbers? and onlyOne #because we can't specify revision numbers for a multi-update
                    #then we should check to see if what we just got meets the requirements
                    matches = ( state.updated.revision[revisionName] is revisionNumber for revisionName, revisionNumber of revisionNumbers )
                    if not matches.every((x)->x)
                        #we have failed to meet the required version number
                        transaction.execution.info.push """
                            #{instructionName} failed to find documents to match the required revisions. 
                            Found: #{JSON.stringify state.updated}. 
                            Required: #{JSON.stringify revisionNumbers}.
                            """
                        return done(null, false)
                    else
                        #revision requirements met, proceed with update
                        #change the updateOps to actually make the revision increments
                        executeUpdate()
                else
                    #we can just go for the update
                    executeUpdate()


_updateOpRollback = (onlyOne, config, transaction, state, collectionName, selector, updateOps, rollbackProjector, revisions, etc, done) ->
    instructionName = if onlyOne then 'updateOneOpRollback' else 'updateManyOpRollback'
    if etc.length isnt 0 then return done(new Error("too many values passed to #{instructionName}"))
    if not (collectionName? and selector? and updateOps?) then return done(new Error("null or missing argument to #{instructionName} command"))

    [error, revisionNames, revisionNumbers] = _processRevisions instructionName, config, collectionName, revisions
    if error?
        return done(error)

    config.db.collection collectionName, {strict: true}, (err, collection) ->
        if err? then return done(err, false)

        if not state.updated?
            #then we never got started on this transaction, so nothing to rollback
            return done(null, true)
        else
            #then we wrote ids and revisions before the update, can we still find these documents?
            #we could easily have failed in the middle of an updateMany, so check one by one.
            docs = if onlyOne then [state.updated] else state.updated

            rollbackDoc = (doc, rollbackDone) ->
                revisionSelector = _id: doc._id
                for revisionName, revisionNumber of doc.revision
                    revisionSelector["revision.#{revisionName}"] = revisionNumber
                collection.count revisionSelector, (err, count) ->
                    if count is 1
                        #then we have not made an update, so there's nothing to rollback
                        return rollbackDone(null, true)
                    else if count is 0
                        #then we must have updated, so we must rollback. This
                        #is true because we wrote our state.updated values
                        #from a within-queue read.
                        mongoRollbackOps = $set: {}, $inc: {}, $unset:{}
                        #only decrement the revision numbers that we specifically asked to increment
                        mongoRollbackOps.$inc["revision.#{revisionName}"] = -1 for revisionName in revisionNames
                        #take the projection that was stored and convert it to mongo dot notation
                        flattened = _mongoFlatten(doc)
                        #the stored projection could contain too many revision keys, so don't copy these. 
                        #also don't update _id.
                        mongoRollbackOps.$set[key] = value for [key, value] in flattened when key.indexOf('revision.') isnt 0
                        #make a list of fields to unset
                        #must include any operation that sets a field which wasnt in state.updated; take care of rename
                        originalDotPaths = ( key for key, value of mongoRollbackOps.$set when key.indexOf('revision.') isnt 0)
                        settyOperators = ['__inc', '__setOnInsert', '__set', '__addToSet', '__push', '__bit']
                        updatedDotPaths = []
                        for op in settyOperators
                            if updateOps[op]?
                                updatedDotPaths.push(path) for path of updateOps[op]
                        unsetDotPaths = ( path for path in updatedDotPaths when path not in originalDotPaths )
                        #$rename is a special case, it's the rename-to field that we're interested
                        if updateOps['__rename']?
                            renamedTo = ( to for from, to of updateOps['__rename'] )
                            unsetDotPaths.push(to) for to in renamedTo when to not in originalDotPaths
                        for path in unsetDotPaths
                            mongoRollbackOps.$unset[path] = 1
                        #remove _id from $set, we dont want to update this
                        delete mongoRollbackOps.$set._id

                        #choose just the document that was updated from it's _id
                        mongoSelector = _id: doc._id

                        return collection.update mongoSelector, mongoRollbackOps, _updateOneOptions, (err, updated) ->
                            if err? then return rollbackDone(err, false)
                            if updated isnt 1
                                transaction.execution.info.push """
                                    updateOneOpRollback failed to update exactly one document (actually #{updated}) during rollback
                                    """
                                return rollbackDone(null, false) #the rollback has failed
                            return rollbackDone(null, true)
                    else
                        #general rollback failure
                        transaction.execution.info.push """
                            updateOneOpRollback found too many documents (#{count}) when trying to find out if an updateOneOp had occurred
                            """
                        return rollbackDone(null, false)
            async.mapSeries docs, rollbackDoc, (err, rollbackResults) ->
                if err? then return done(err, null)
                #rollback is a success if each individual rollback returns status true.
                return done(err, rollbackResults.every((x) -> x))


# should be possible to do committed.register('db', committed.db) to use these fns
#note: values passed to instructions should generally be considered immutable,
#as they'll be cloned after commit in the event of implicit rollback
#instructions must return done(err, result) where result is true for success and false for failure

# a revision is a counter, incremented during each update. Prior to each
# update the id and revision number of each document that'll be changed is
# recorded in the instruction state. Warning: for updates that effect millions
# of documents, this will result in millions of ids in the transaction.

# revisioning needs to be explicitly setup during start, as config.revision, eg
# { collectionName1: [subSchema1, subSchema2] } would be interpreted as:
#  - collectionName1 had two sub-schemas, so at insert collection1Doc.revision is {subSchema1: 0 subSchema2: 0}
#  - collectionName2 is not a revisioned collection and so will not have a .revision property.

# if you decide to fail an instruction by returning done(null, false) you should push a reason to transaction.execution.info

exports.db =
    pass: (config, transaction, state, args, done) ->
        done(null, true)

    # an insert instruction. documents can be a single document or an array
    # thereof. revisions is optional, if present should be an array containing
    # the subschemas which are to be established in each document's .revision
    # subdoc. revisions may only contain subschemas mentioned in the
    # config.revisions.collectionName object. 
    insert: (config, transaction, state, [collectionName, documents, revisions, etc...], done) ->
        if etc.length isnt 0 then return done(new Error("too many values passed to insert"))
        if not (collectionName? and documents?) then return done(new Error("null or missing argument to insert command"))
        if not Array.isArray(documents) then documents = [documents]

        [error, revisionNames, revisionNumbers] = _processRevisions 'insert', config, collectionName, revisions
        if error?
            return done(error)
        if revisionNumbers? then return done(new Error("can't insert using specific revision numbers #{revisionNumbers}"))

        #check the docs dont have any revision info if we're intending to automatically set revision info
        if documents.some((d) -> d.revision?)
            return done(new Error("can't insert documents when some already have a revision property"))

        config.db.collection collectionName, {strict: true}, (err, collection) ->
            if err? then return done(err, false)
            #inserts should _id every document, otherwise accurate rollback won't be possible, moreover these ids need to be saved somewhere
            state.ids = []
            for d in documents
                if d._id? 
                    state.ids.push d._id
                else
                    d._id = _pkFactory.createPk()
                    state.ids.push d._id
                #while we're looking at documents, check to see if we should create some revision info
                if revisionNames.length > 0 and not d.revision?
                    d.revision = {}
                    for revisionName in revisionNames
                        d.revision[revisionName] = 0
            options =
                w:1
                journal: true
                serializeFunctions: false
            _updateTransactionState transaction, (err) ->
                if err? then return done(err, null)
                collection.insert documents, options, (err, objects) ->
                    if err? then return done(err, null)
                    return done(null, true)

    insertRollback: (config, transaction, state, [collectionName, documents, revisions, etc...], done) ->
        if etc.length isnt 0 then return done(new Error("too many values passed to insertRollback"))
        if not (collectionName? and documents?) then return done(new Error("null or missing argument to insertRollback command"))

        config.db.collection collectionName, {strict: true}, (err, collection) ->
            if err? then return done(err, false)
            options =
                w:1
                journal: true
                serializeFunctions: false
            if not state.ids?
                #then we never got to the insert, so there's nothing for the rollback to do
                return done(null, true)
            else
                #where documents lack an _id, we have to fill this in from what was generated during the insert instruction. 
                if not Array.isArray(documents) then documents = [documents]
                if documents.length isnt state.ids.length
                    return done(new Error("during rollback, number of documents to insert doesnt match the number of ids generated during insert"), false)
                for d,i in documents
                    if not d._id? then d._id = state.ids[i]
                iterator = (item, itemDone) ->
                    #remove justOne item, but only if it's an exact match for the inserted document
                    collection.remove item, options, (err, result) ->
                        itemDone(err)
                return async.each documents, iterator, (err) ->
                    return done(err, true) #rollbacks only fail if there's an error


    #updateOneOp is an update operation on one document using mongo
    #operations. revisions is optional, if  present then it needs to be either 
    # 1. a string containing revisions names which we should increment
    # 2. an array of such strings
    # 3. an object whose key,value are a revision name and revision number respectively
    # if you supply revision numbers, that saves a read+write.
    # mongoProjector is used to take a copy of the pre-commit values, if it's null, the whole document will be copied
    # the object must not contain projection operations.
    updateOneOp: (config, transaction, state, [collectionName, selector, updateOps, rollbackProjector, revisions, etc...], done) ->
        return _updateOpInstruction(true, config, transaction, state, collectionName, selector, updateOps, rollbackProjector, revisions, etc, done)

    updateOneOpRollback: (config, transaction, state, [collectionName, selector, updateOps, rollbackProjector, revisions, etc...], done) ->
        return _updateOpRollback(true, config, transaction, state, collectionName, selector, updateOps, rollbackProjector, revisions, etc, done)

    updateManyOp: (config, transaction, state, [collectionName, selector, updateOps, rollbackProjector, revisions, etc...], done) ->
        return _updateOpInstruction(false, config, transaction, state, collectionName, selector, updateOps, rollbackProjector, revisions, etc, done)

    updateManyOpRollback: (config, transaction, state, [collectionName, selector, updateOps, rollbackProjector, revisions, etc...], done) ->
        return _updateOpRollback(false, config, transaction, state, collectionName, selector, updateOps, rollbackProjector, revisions, etc, done)

    #updateOneDoc updates a parts of a document. It isnt save, if you want to
    #save the whole document you should use a different instruction.

    updateOneDoc: (config, transaction, state, [collectionName, newPartialDocument, oldPartialDocument, etc...], done) ->
        if etc.length isnt 0 then return done(new Error("too many values passed to updateOneDoc"))
        if not (collectionName? and newPartialDocument? and oldPartialDocument?) then return done(new Error("null or missing argument to updateOneDoc command"))
        if not oldPartialDocument._id? or not newPartialDocument._id? or not oldPartialDocument._id.equals(newPartialDocument._id)
            return done(new Error("both old and new partial documents must have _ids, and they must match to updateOneDoc"))
        if not oldPartialDocument.revision? then return done(new Error("oldPartialDocument must contain revision information to updateOneDoc"))

        #take the revision sub-doc of the old partial doc to be the revisions that we're interested in
        revisionNames = ( revisionName for revisionName of oldPartialDocument.revision )

        [error, x, y] = _processRevisions 'updateOneDoc', config, collectionName, revisionNames
        if error?
            return done(error)

        #first we must establish whether the revision requirements for this
        #command can be satisfied. specifically, whether the document revision
        #in oldPartialDocument is current
        config.db.collection collectionName, {strict: true}, (err, collection) ->
            if err? then return done(err, false)
            oldSelector = {}
            oldSelector[key] = value for [key, value] in _mongoFlatten(oldPartialDocument)
            collection.count oldSelector, (err, count) ->
                if err? then return done(err, false)
                #this will also confirm that in the event of rollback we'll be writing something valid.
                if count isnt 1
                    transaction.execution.info.push """
                        updateOneDoc failed to find exactly one document (found #{count}) matching the given oldPartialDocument using #{oldSelector})
                        """
                    return done(null, false)
                #if we find exactly one match then our revision requirement is met, we should record the fact that the instruction is safe to proceed
                state.safeToExecuteInstruction = true
                _updateTransactionState transaction, (err) ->
                    if err? then return done(err, null)
                    #the state is now saved - we'll know what to do in the event of rollback.
                    #now work out the operations necessary to do a partialdoc update
                    mongoOps = $set: {}, $unset:{}, $inc:{}
                    _mongoOpsFromTo(mongoOps, oldPartialDocument, newPartialDocument)
                    #update all the relevant revisions
                    mongoOps.$inc["revision.#{revisionName}"] = 1 for revisionName of oldPartialDocument.revision
                    collection.update {_id: oldPartialDocument._id}, mongoOps, _updateOneOptions, (err, updated) ->
                        if err? then return done(err, null)
                        if updated isnt 1 
                            transaction.execution.info.push """
                                updateOneDoc must update exactly one document, using #{JSON.stringify mongoOps} #{updated} were updated
                                """
                            return done(null, false) #the instruction has failed
                        return done(null, true)


    updateOneDocRollback: (config, transaction, state, [collectionName, newPartialDocument, oldPartialDocument, etc...], done) ->
        if etc.length isnt 0 then return done(new Error("too many values passed to updateOneDocRollback"))
        if not (collectionName? and newPartialDocument? and oldPartialDocument?) then return done(new Error("null or missing argument to updateOneDocRollback command"))
        if not oldPartialDocument._id? or not newPartialDocument._id? or not oldPartialDocument._id.equals(newPartialDocument._id)
            return done(new Error("both old and new partial documents must have _ids, and they must match to updateOneDocRollback"))
        if not oldPartialDocument.revision? then return done(new Error("oldPartialDocument must contain revision information to updateOneDocRollback"))

        if not state.safeToExecuteInstruction
            #then the revision requierments for this instruction were not met,
            #or the instruction was interrupted early, so there is nothing to
            #rollback
            return done(null, true)
        else
            #revision requirments for the original instruction were met
            config.db.collection collectionName, {strict: true}, (err, collection) ->
                if err? then return done(err, false)

                #because revision requirements were met at start of
                #instruction, if the revision numbers was bumped up, it means
                #the instruction was executed
                selector = {_id: oldPartialDocument._id }
                selector["revision.#{revisionName}"] = oldPartialDocument.revision[revisionName] + 1 for revisionName of oldPartialDocument.revision
                collection.count selector, (err, count) ->
                    if err? then return done(err, false)
                    if count is 0
                        #then the update was never completed, there's nothing to rollback
                        return done(null, true)
                    else
                        #the update must have gone through, rollback
                        mongoOps = $set: {}, $unset:{}
                        _mongoOpsFromTo(mongoOps, newPartialDocument, oldPartialDocument)
                        #reset all the relevant revisions
                        mongoOps.$set["revision.#{revisionName}"] = oldPartialDocument.revision[revisionName] for revisionName of oldPartialDocument.revision

                        collection.update {_id: oldPartialDocument._id}, mongoOps, _updateOneOptions, (err, updated) ->
                                if err? then return done(err, null)
                                if updated isnt 1 
                                    transaction.execution.info.push """
                                        updateOneDocRollback must update exactly one document, using #{JSON.stringify mongoOps} #{updated} were updated
                                        """
                                    return done(null, false) #the instruction has failed
                                return done(null, true)
